@@ -31,9 +31,12 @@
 *     [12] Guha, Sudipto, et al.
 *          "Robust random cut forest based anomaly detection on streams."
 *          International conference on machine learning. PMLR, 2016.
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 * 
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -91,6 +94,10 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
             goto terminal_statistics;
     }
 
+    /* if there's no columns left to split, can end here */
+    if (!workspace.col_sampler.get_remaining_cols())
+        goto terminal_statistics;
+
     /* for sparse matrices, need to sort the indices */
     if (input_data.Xc_indptr != NULL && impute_nodes == NULL)
         std::sort(workspace.ix_arr.begin() + workspace.st, workspace.ix_arr.begin() + workspace.end + 1);
@@ -834,6 +841,16 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
         /* add another round of separation depth for distance */
         if (model_params.calc_dist && curr_depth > 0)
             add_separation_step(workspace, input_data, (double)(-1));
+
+        /* if it split by a categorical variable with only 2 values,
+           the column will no longer be splittable in either branch */
+        if (trees.back().col_type == Categorical &&
+            ((model_params.cat_split_type == SubSet && trees.back().cat_split.empty()) ||
+             (model_params.cat_split_type == SingleCateg && input_data.ncat[trees.back().col_num] == 2)))
+        {
+            workspace.col_sampler.drop_col(trees.back().col_num + input_data.ncols_numeric,
+                                           workspace.end - workspace.st + 1);
+        }
         
         size_t tree_from = trees.size() - 1;
         std::unique_ptr<RecursionState>
@@ -1074,6 +1091,11 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                     }
                     break;
                 }
+
+                default:
+                {
+                    assert(0);
+                }
             }
         }
 
@@ -1091,6 +1113,17 @@ void split_itree_recursive(std::vector<IsoTree>     &trees,
                     trees.back().cat_split[cat] = new_to_left;
         }
 
+        /* If doing single-category splits, the branch that got only one category will not
+           be splittable anymore, so it can be dropped for the remainder of that branch */
+        if (trees.back().col_type == Categorical &&
+            model_params.cat_split_type == SingleCateg &&
+            input_data.ncat[trees.back().col_num] > 2 /* <- in this case, would have been dropped earlier */
+            )
+        {
+            workspace.col_sampler.drop_col(trees.back().col_num + input_data.ncols_numeric,
+                                           workspace.end - workspace.st + 1);
+        }
+
         /* left branch */
         trees.back().tree_left = trees.size();
         trees.emplace_back();
