@@ -31,9 +31,12 @@
 *     [12] Guha, Sudipto, et al.
 *          "Robust random cut forest based anomaly detection on streams."
 *          International conference on machine learning. PMLR, 2016.
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 * 
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -501,6 +504,7 @@ public:
     int n_present;
     int n_left;
     void initialize(size_t max_depth, int max_categ, bool reserve_counts, ScoringMetric scoring_metric);
+    [[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
     void push_density(double xmin, double xmax, double split_point);
     void push_density(size_t counts[], int ncat);
     void push_density(int n_left, int n_present);
@@ -511,6 +515,7 @@ public:
     void push_adj(size_t *restrict counts, int ncat, int chosen_cat, ScoringMetric scoring_metric);
     void push_adj(double pct_tree_left, ScoringMetric scoring_metric);
     void pop();
+    [[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
     double calc_density(long double remainder, size_t sample_size);
     ldouble_safe calc_adj_depth();
     double calc_adj_density();
@@ -551,6 +556,7 @@ public:
     bool sample_col(size_t &col, RNG_engine &rnd_generator);
     void prepare_full_pass();        /* when passing through all columns */
     bool sample_col(size_t &col); /* when passing through all columns */
+    void drop_col(size_t col, size_t nobs_left);
     void drop_col(size_t col);
     void shuffle_remainder(RNG_engine &rnd_generator);
     bool has_weights();
@@ -801,19 +807,21 @@ void simplify_hplane(IsoHPlane &hplane, WorkerMemory &workspace, InputData &inpu
 
 /* predict.cpp */
 template <class real_t, class sparse_ix>
-void predict_iforest(real_t numeric_data[], int categ_data[],
-                     bool is_col_major, size_t ncols_numeric, size_t ncols_categ,
-                     real_t Xc[], sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
-                     real_t Xr[], sparse_ix Xr_ind[], sparse_ix Xr_indptr[],
+[[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno"), gnu::hot]]
+void predict_iforest(real_t *restrict numeric_data, int *restrict categ_data,
+                     bool is_col_major, size_t ld_numeric, size_t ld_categ,
+                     real_t *restrict Xc, sparse_ix *restrict Xc_ind, sparse_ix *restrict Xc_indptr,
+                     real_t *restrict Xr, sparse_ix *restrict Xr_ind, sparse_ix *restrict Xr_indptr,
                      size_t nrows, int nthreads, bool standardize,
                      IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
-                     double output_depths[],   sparse_ix tree_num[],
-                     double per_tree_depths[]);
+                     double *restrict output_depths,   sparse_ix *restrict tree_num,
+                     double *restrict per_tree_depths);
 template <class PredictionData, class sparse_ix>
+[[gnu::hot]]
 void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
                                IsoForest             &model_outputs,
                                PredictionData        &prediction_data,
-                               double                &output_depth,
+                               double &restrict      output_depth,
                                sparse_ix *restrict   tree_num,
                                double *restrict      tree_depth,
                                size_t                row);
@@ -832,7 +840,7 @@ template <class PredictionData, class sparse_ix>
 void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
                           ExtIsoForest            &model_outputs,
                           PredictionData          &prediction_data,
-                          double                  &output_depth,
+                          double &restrict        output_depth,
                           sparse_ix *restrict     tree_num,
                           double *restrict        tree_depth,
                           size_t                  row);
@@ -840,7 +848,7 @@ template <class PredictionData, class sparse_ix, class ImputedData>
 void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                      ExtIsoForest             &model_outputs,
                      PredictionData           &prediction_data,
-                     double                   &output_depth,
+                     double &restrict         output_depth,
                      std::vector<ImputeNode> *impute_nodes,
                      ImputedData             *imputed_data,
                      sparse_ix *restrict      tree_num,
@@ -849,15 +857,15 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
 template <class real_t, class sparse_ix>
 void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int nthreads,
                          IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
-                         double output_depths[],   sparse_ix tree_num[],
-                         double per_tree_depths[]);
+                         double *restrict output_depths,   sparse_ix *restrict tree_num,
+                         double *restrict per_tree_depths);
 template <class PredictionData, class sparse_ix>
 void traverse_itree_csc(WorkerForPredictCSC   &workspace,
                         std::vector<IsoTree>  &trees,
                         IsoForest             &model_outputs,
                         PredictionData        &prediction_data,
-                        sparse_ix             *tree_num,
-                        double                *per_tree_depths,
+                        sparse_ix *restrict   tree_num,
+                        double *restrict      per_tree_depths,
                         size_t                curr_tree,
                         bool                  has_range_penalty);
 template <class PredictionData, class sparse_ix>
@@ -865,14 +873,14 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
                          std::vector<IsoHPlane>   &hplanes,
                          ExtIsoForest             &model_outputs,
                          PredictionData           &prediction_data,
-                         sparse_ix                *tree_num,
-                         double                   *per_tree_depths,
+                         sparse_ix *restrict      tree_num,
+                         double *restrict         per_tree_depths,
                          size_t                   curr_tree,
                          bool                     has_range_penalty);
 template <class PredictionData>
 void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
                            PredictionData       &prediction_data,
-                           double               *weights_arr,
+                           double *restrict     weights_arr,
                            size_t               col_num,
                            double               range_low,
                            double               range_high);
@@ -1030,10 +1038,12 @@ void calc_kurt_all_cols(InputData &input_data, WorkerMemory &workspace, ModelPar
 
 /* utils.cpp */
 size_t log2ceil(size_t x);
+[[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
 double digamma(double x);
 double harmonic(size_t n);
 double harmonic_recursive(double a, double b);
 double expected_avg_depth(size_t sample_size);
+[[gnu::optimize("no-trapping-math"), gnu::optimize("no-math-errno")]]
 double expected_avg_depth(long double approx_sample_size);
 double expected_separation_depth(size_t n);
 double expected_separation_depth_hotstart(double curr, size_t n_curr, size_t n_final);
