@@ -31,9 +31,12 @@
 *     [12] Guha, Sudipto, et al.
 *          "Robust random cut forest based anomaly detection on streams."
 *          International conference on machine learning. PMLR, 2016.
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 * 
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -158,8 +161,7 @@
 * - output_depths[nrows] (out)
 *       Pointer to array where the output average depths or outlier scores will be written into
 *       (the return type is controlled according to parameter 'standardize').
-*       Must already be initialized to zeros, and should always be passed when calling
-*       this function (it is not optional).
+*       Should always be passed when calling this function (it is not optional).
 * - tree_num[nrows * ntrees] (out)
 *       Pointer to array where the output terminal node numbers will be written into.
 *       Note that the mapping between tree node and terminal tree node is not stored in
@@ -178,14 +180,14 @@
 *       Pass NULL if this type of output is not needed.
 */
 template <class real_t, class sparse_ix>
-void predict_iforest(real_t numeric_data[], int categ_data[],
+void predict_iforest(real_t *restrict numeric_data, int *restrict categ_data,
                      bool is_col_major, size_t ld_numeric, size_t ld_categ,
-                     real_t Xc[], sparse_ix Xc_ind[], sparse_ix Xc_indptr[],
-                     real_t Xr[], sparse_ix Xr_ind[], sparse_ix Xr_indptr[],
+                     real_t *restrict Xc, sparse_ix *restrict Xc_ind, sparse_ix *restrict Xc_indptr,
+                     real_t *restrict Xr, sparse_ix *restrict Xr_ind, sparse_ix *restrict Xr_indptr,
                      size_t nrows, int nthreads, bool standardize,
                      IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
-                     double output_depths[],   sparse_ix tree_num[],
-                     double per_tree_depths[])
+                     double *restrict output_depths,   sparse_ix *restrict tree_num,
+                     double *restrict per_tree_depths)
 {
     /* put data in a struct for passing it in fewer lines */
     PredictionData<real_t, sparse_ix>
@@ -221,17 +223,19 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
                     shared(nrows, model_outputs, prediction_data, output_depths, tree_num, per_tree_depths)
             for (size_t_for row = 0; row < (decltype(row))nrows; row++)
             {
+                double score = 0;
                 for (size_t tree = 0; tree < model_outputs->trees.size(); tree++)
                 {
                     traverse_itree_no_recurse(model_outputs->trees[tree],
                                               *model_outputs,
                                               prediction_data,
-                                              output_depths[row],
+                                              score,
                                               (tree_num == NULL)? NULL : (tree_num + nrows * tree),
                                               (per_tree_depths == NULL)?
                                                   NULL : (per_tree_depths + tree + row*model_outputs->trees.size()),
                                               (size_t) row);
                 }
+                output_depths[row] = score;
             }
         }
 
@@ -241,20 +245,22 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
                     shared(nrows, model_outputs, prediction_data, output_depths, tree_num, per_tree_depths)
             for (size_t_for row = 0; row < (decltype(row))nrows; row++)
             {
+                double score = 0;
                 for (size_t tree = 0; tree < model_outputs->trees.size(); tree++)
                 {
-                    output_depths[row] += traverse_itree(model_outputs->trees[tree],
-                                                         *model_outputs,
-                                                         prediction_data,
-                                                         (std::vector<ImputeNode>*)NULL,
-                                                         (ImputedData<sparse_ix>*)NULL,
-                                                         (double)0,
-                                                         (size_t) row,
-                                                         (tree_num == NULL)? NULL : (tree_num + nrows * tree),
-                                                         (per_tree_depths == NULL)?
-                                                            NULL : (per_tree_depths + tree + row*model_outputs->trees.size()),
-                                                         (size_t) 0);
+                    score += traverse_itree(model_outputs->trees[tree],
+                                            *model_outputs,
+                                            prediction_data,
+                                            (std::vector<ImputeNode>*)NULL,
+                                            (ImputedData<sparse_ix>*)NULL,
+                                            (double)0,
+                                            (size_t) row,
+                                            (tree_num == NULL)? NULL : (tree_num + nrows * tree),
+                                            (per_tree_depths == NULL)?
+                                                NULL : (per_tree_depths + tree + row*model_outputs->trees.size()),
+                                            (size_t) 0);
                 }
+                output_depths[row] = score;
             }
         }
     }
@@ -274,17 +280,19 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
                     shared(nrows, model_outputs_ext, prediction_data, output_depths, tree_num, per_tree_depths)
             for (size_t_for row = 0; row < (decltype(row))nrows; row++)
             {
+                double score = 0;
                 for (size_t tree = 0; tree < model_outputs_ext->hplanes.size(); tree++)
                 {
                     traverse_hplane_fast(model_outputs_ext->hplanes[tree],
                                          *model_outputs_ext,
                                          prediction_data,
-                                         output_depths[row],
+                                         score,
                                          (tree_num == NULL)? NULL : (tree_num + nrows * tree),
                                          (per_tree_depths == NULL)?
                                             NULL : (per_tree_depths + tree + row*model_outputs_ext->hplanes.size()),
                                          (size_t) row);
                 }
+                output_depths[row] = score;
             }
         }
 
@@ -294,12 +302,13 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
                     shared(nrows, model_outputs_ext, prediction_data, output_depths, tree_num, per_tree_depths)
             for (size_t_for row = 0; row < (decltype(row))nrows; row++)
             {
+                double score = 0;
                 for (size_t tree = 0; tree < model_outputs_ext->hplanes.size(); tree++)
                 {
                     traverse_hplane(model_outputs_ext->hplanes[tree],
                                     *model_outputs_ext,
                                     prediction_data,
-                                    output_depths[row],
+                                    score,
                                     (std::vector<ImputeNode>*)NULL,
                                     (ImputedData<sparse_ix>*)NULL,
                                     (tree_num == NULL)? NULL : (tree_num + nrows * tree),
@@ -307,6 +316,7 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
                                         NULL : (per_tree_depths + tree + row*model_outputs_ext->hplanes.size()),
                                     (size_t) row);
                 }
+                output_depths[row] = score;
             }
         }
     }
@@ -326,8 +336,13 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
     }
 
     if (standardize)
+        #ifdef _WIN32
         #pragma omp parallel for if(nrows > 100) schedule(static) num_threads(nthreads) \
                 shared(nrows, output_depths, depth_divisor)
+        #else
+        #pragma omp parallel for simd if(nrows > 100) schedule(static) num_threads(nthreads) \
+                shared(nrows, output_depths, depth_divisor)
+        #endif
         for (size_t_for row = 0; row < (decltype(row))nrows; row++)
             output_depths[row] = std::exp2( - output_depths[row] / depth_divisor );
     else
@@ -345,11 +360,10 @@ void predict_iforest(real_t numeric_data[], int categ_data[],
 }
 
 template <class PredictionData, class sparse_ix>
-[[gnu::hot]]
 void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
                                IsoForest             &model_outputs,
                                PredictionData        &prediction_data,
-                               double                &output_depth,
+                               double &restrict      output_depth,
                                sparse_ix *restrict   tree_num,
                                double *restrict      tree_depth,
                                size_t                row)
@@ -475,7 +489,6 @@ void traverse_itree_no_recurse(std::vector<IsoTree>  &tree,
 enum NumericConfig {DenseRowMajor, DenseColMajor, SparseCSR, SparseCSC};
 
 template <class PredictionData, class sparse_ix, class ImputedData>
-[[gnu::hot]]
 double traverse_itree(std::vector<IsoTree>     &tree,
                       IsoForest                &model_outputs,
                       PredictionData           &prediction_data,
@@ -766,11 +779,10 @@ double traverse_itree(std::vector<IsoTree>     &tree,
 /* this is a simpler version for situations in which there is
    only numeric data in dense arrays, no missing values, no range penalty */
 template <class PredictionData, class sparse_ix>
-[[gnu::hot]]
 void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
                           ExtIsoForest            &model_outputs,
                           PredictionData          &prediction_data,
-                          double                  &output_depth,
+                          double &restrict        output_depth,
                           sparse_ix *restrict     tree_num,
                           double *restrict        tree_depth,
                           size_t                  row)
@@ -815,11 +827,10 @@ void traverse_hplane_fast(std::vector<IsoHPlane>  &hplane,
 
 /* this is the full version that works with potentially missing values, sparse matrices, and categoricals */
 template <class PredictionData, class sparse_ix, class ImputedData>
-[[gnu::hot]]
 void traverse_hplane(std::vector<IsoHPlane>   &hplane,
                      ExtIsoForest             &model_outputs,
                      PredictionData           &prediction_data,
-                     double                   &output_depth,
+                     double &restrict         output_depth,
                      std::vector<ImputeNode> *impute_nodes,     /* only when imputing missing */
                      ImputedData             *imputed_data,     /* only when imputing missing */
                      sparse_ix *restrict      tree_num,
@@ -1006,8 +1017,8 @@ void traverse_hplane(std::vector<IsoHPlane>   &hplane,
 template <class real_t, class sparse_ix>
 void batched_csc_predict(PredictionData<real_t, sparse_ix> &prediction_data, int nthreads,
                          IsoForest *model_outputs, ExtIsoForest *model_outputs_ext,
-                         double output_depths[],   sparse_ix tree_num[],
-                         double per_tree_depths[])
+                         double *restrict output_depths,   sparse_ix *restrict tree_num,
+                         double *restrict per_tree_depths)
 {
     #ifdef _OPENMP
     size_t ntrees = (model_outputs != NULL)? model_outputs->trees.size() : model_outputs_ext->hplanes.size();
@@ -1156,8 +1167,8 @@ void traverse_itree_csc(WorkerForPredictCSC   &workspace,
                         std::vector<IsoTree>  &trees,
                         IsoForest             &model_outputs,
                         PredictionData        &prediction_data,
-                        sparse_ix             *tree_num,
-                        double                *per_tree_depths,
+                        sparse_ix *restrict   tree_num,
+                        double *restrict      per_tree_depths,
                         size_t                curr_tree,
                         bool                  has_range_penalty)
 {
@@ -1400,8 +1411,8 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
                          std::vector<IsoHPlane>   &hplanes,
                          ExtIsoForest             &model_outputs,
                          PredictionData           &prediction_data,
-                         sparse_ix                *tree_num,
-                         double                   *per_tree_depths,
+                         sparse_ix *restrict      tree_num,
+                         double *restrict         per_tree_depths,
                          size_t                   curr_tree,
                          bool                     has_range_penalty)
 {
@@ -1522,7 +1533,7 @@ void traverse_hplane_csc(WorkerForPredictCSC      &workspace,
 template <class PredictionData>
 void add_csc_range_penalty(WorkerForPredictCSC  &workspace,
                            PredictionData       &prediction_data,
-                           double               *weights_arr,
+                           double *restrict     weights_arr,
                            size_t               col_num,
                            double               range_low,
                            double               range_high)
