@@ -31,9 +31,12 @@
 *     [12] Guha, Sudipto, et al.
 *          "Robust random cut forest based anomaly detection on streams."
 *          International conference on machine learning. PMLR, 2016.
+*     [13] Cortes, David.
+*          "Isolation forests: looking beyond tree depth."
+*          arXiv preprint arXiv:2111.11639 (2021).
 * 
 *     BSD 2-Clause License
-*     Copyright (c) 2019-2021, David Cortes
+*     Copyright (c) 2019-2022, David Cortes
 *     All rights reserved.
 *     Redistribution and use in source and binary forms, with or without
 *     modification, are permitted provided that the following conditions are met:
@@ -59,7 +62,7 @@
    https://stackoverflow.com/questions/2589096/find-most-significant-bit-left-most-that-is-set-in-a-bit-array
    https://stackoverflow.com/questions/11376288/fast-computing-of-log2-for-64-bit-integers  */
 #if SIZE_MAX == UINT32_MAX /* 32-bit systems */
-    constexpr static const int MultiplyDeBruijnBitPosition[32] =
+    constexpr static const uint32_t MultiplyDeBruijnBitPosition[32] =
     {
         0, 9,  1,  10, 13, 21, 2, 29, 11, 14, 16, 18, 22, 25, 3, 30,
         8, 12, 20, 28, 15, 17, 24, 7, 19, 27, 23, 6,  26, 5,  4, 31
@@ -149,7 +152,7 @@ double digamma(double x)
         y = 0.0;
     }
 
-    y = std::log(x)  -  (0.5/x)  -  y;
+    y = (-0.5/x) + std::log(x) - y;
     return y;
 }
 
@@ -159,13 +162,17 @@ double digamma(double x)
 double harmonic(size_t n)
 {
     if (n > THRESHOLD_EXACT_H)
-        return std::log((long double)n) + (long double)EULERS_GAMMA
-                + 0.5 * (1./(long double)n)
-                - 0.5 * (1./square((long double)n))
-                      * ( 1./6. -   (1./square((long double)n))
-                                  * (1./60. - (1./126.)*(1./square((long double)n))) );
+    {
+        long double temp = 1.0l / square((long double)n);
+        return  - 0.5l * temp * ( 1.0l/6.0l  -   temp * (1.0l/60.0l - (1.0l/126.0l)*temp) )
+                + 0.5l * (1.0l/(long double)n)
+                + std::log((long double)n) + (long double)EULERS_GAMMA;
+    }
+    
     else
+    {
         return harmonic_recursive((double)1, (double)(n + 1));
+    }
 }
 
 double harmonic_recursive(double a, double b)
@@ -204,12 +211,12 @@ double expected_avg_depth(long double approx_sample_size)
         return 0;
     else if (approx_sample_size < (long double)INT32_MAX)
         return 2. * (digamma(approx_sample_size + 1.) + EULERS_GAMMA - 1.);
-    else
-        return 2. * std::log(approx_sample_size) + 2.*((long double)EULERS_GAMMA - 1.)
-                + (1./approx_sample_size)
-                - (1./square(approx_sample_size))
-                   * ( 1./6. -   (1./square(approx_sample_size))
-                               * (1./60. - (1./126.)*(1./square(approx_sample_size))) );
+    else {
+        long double temp = 1.0l / square(approx_sample_size);
+        return 2.0l * std::log(approx_sample_size) + 2.0l*((long double)EULERS_GAMMA - 1.0l)
+               + (1.0l/approx_sample_size)
+               - temp * ( 1.0l/6.0l -   temp * (1.0l/60.0l - (1.0l/126.0l)*temp) );
+    }
 }
 
 /* https://math.stackexchange.com/questions/3388518/expected-number-of-paths-required-to-separate-elements-in-a-binary-tree */
@@ -853,11 +860,32 @@ void ColumnSampler::leave_m_cols(size_t m, RNG_engine &rnd_generator)
     }
 }
 
-void ColumnSampler::drop_col(size_t col)
+void ColumnSampler::drop_col(size_t col, size_t nobs_left)
 {
     if (!this->has_weights())
     {
-        std::swap(this->col_indices[this->last_given], this->col_indices[--this->curr_pos]);
+        if (this->col_indices[this->last_given] == col)
+        {
+            std::swap(this->col_indices[this->last_given], this->col_indices[--this->curr_pos]);
+        }
+
+        else if (this->curr_pos > 4*nobs_left)
+        {
+            return;
+        }
+
+        else
+        {
+            for (size_t ix = 0; ix < this->curr_pos; ix++)
+            {
+                if (this->col_indices[ix] == col)
+                {
+                    std::swap(this->col_indices[ix], this->col_indices[--this->curr_pos]);
+                    break;
+                }
+            }
+        }
+
         if (this->curr_col) this->curr_col--;
     }
 
@@ -875,6 +903,11 @@ void ColumnSampler::drop_col(size_t col)
     }
 }
 
+void ColumnSampler::drop_col(size_t col)
+{
+    this->drop_col(col, SIZE_MAX);
+}
+
 void ColumnSampler::prepare_full_pass()
 {
     this->curr_col = 0;
@@ -1326,23 +1359,26 @@ void DensityCalculator::initialize(size_t max_depth, int max_categ, bool reserve
 
 void DensityCalculator::push_density(double xmin, double xmax, double split_point)
 {
-    if (std::isinf(xmax) || std::isinf(xmin))
+    if (std::isinf(xmax) || std::isinf(xmin) || std::isnan(xmin) || std::isnan(xmax) || std::isnan(split_point))
     {
         this->multipliers.push_back(0);
         return;
     }
 
-    double range = xmax - xmin;
-    if (range <= 0) range = std::numeric_limits<double>::min();
-    double mult_left = std::log((split_point - xmin) / range);
-    double mult_right = std::log((xmax - split_point) / range);
-    if (std::isinf(mult_left))
+    double range = std::fmax(xmax - xmin, std::numeric_limits<double>::min());
+    double dleft = std::fmax(split_point - xmin, std::numeric_limits<double>::min());
+    double dright = std::fmax(xmax - split_point, std::numeric_limits<double>::min());
+    double mult_left = std::log(dleft / range);
+    double mult_right = std::log(dright / range);
+    while (std::isinf(mult_left))
     {
-        double diff = split_point - xmin;
-        do {
-            diff = std::nextafter(diff, HUGE_VAL);
-            mult_left = std::log(diff / range);
-        } while (std::isinf(mult_left));
+        dleft = std::nextafter(dleft, (mult_left < 0)? HUGE_VAL : (-HUGE_VAL));
+        mult_left = std::log(dleft / range);
+    }
+    while (std::isinf(mult_right))
+    {
+        dright = std::nextafter(dright, (mult_right < 0)? HUGE_VAL : (-HUGE_VAL));
+        mult_right = std::log(dright / range);
     }
 
     mult_left = std::isnan(mult_left)? 0 : mult_left;
@@ -1363,7 +1399,7 @@ void DensityCalculator::push_density(size_t counts[], int ncat)
 {
     /* this one assumes 'categ_present' has entries 0/1 for missing/present */
     int n_present = 0;
-    for (size_t cat = 0; cat < ncat; cat++)
+    for (int cat = 0; cat < ncat; cat++)
         n_present += counts[cat] > 0;
     this->push_density(0., (double)n_present, 1.);
 }
@@ -1382,27 +1418,18 @@ void DensityCalculator::push_density()
 
 void DensityCalculator::push_adj(double xmin, double xmax, double split_point, double pct_tree_left, ScoringMetric scoring_metric)
 {
-    double range = xmax - xmin;
-    double chunk_left = (split_point - xmin) / range;
-    double chunk_right = (xmax - split_point) / range;
-    if (std::isinf(xmax) || std::isinf(xmin))
+    double range = std::fmax(xmax - xmin, std::numeric_limits<double>::min());
+    double dleft = std::fmax(split_point - xmin, std::numeric_limits<double>::min());
+    double dright = std::fmax(xmax - split_point, std::numeric_limits<double>::min());
+    double chunk_left = dleft / range;
+    double chunk_right = dright / range;
+    if (std::isinf(xmax) || std::isinf(xmin) || std::isnan(xmin) || std::isnan(xmax) || std::isnan(split_point))
     {
         chunk_left = pct_tree_left;
         chunk_right = 1. - pct_tree_left;
         goto add_chunks;
     }
 
-    if (range <= 0) range = std::numeric_limits<double>::min();
-
-    if (!std::isnan(chunk_left) && chunk_left <= 0)
-    {
-        double diff = split_point - xmin;
-        do {
-            diff = std::nextafter(diff, HUGE_VAL);
-            chunk_left = diff / range;
-        } while (chunk_left == 0);
-    }
-
     if (std::isnan(chunk_left) || std::isnan(chunk_right))
     {
         chunk_left = 0.5;
@@ -1486,10 +1513,19 @@ void DensityCalculator::pop()
 double DensityCalculator::calc_density(long double remainder, size_t sample_size)
 {
     long double res = std::exp( std::log(remainder) - std::log((long double)sample_size) - this->multipliers.back() );
+    res = std::fmax(res, (long double)std::numeric_limits<double>::min());
     return std::fmin(res, std::log2((long double)sample_size));
     // return std::fmin(res, expected_avg_depth(sample_size));
     // return res;
     // return std::fmin(res, 2.0l*std::log2((long double)sample_size));
+    // if (res >= HUGE_VAL)
+    // {
+    //     res = HUGE_VAL;
+    //     do
+    //     {
+    //         res = std::nextafter(res, 0.0l);
+    //     } while ((double)res >= HUGE_VAL);
+    // }
 }
 
 ldouble_safe DensityCalculator::calc_adj_depth()
@@ -2152,7 +2188,7 @@ void divide_subset_split(size_t *restrict ix_arr, int x[], size_t st, size_t end
     /* if NAs are not to be bothered with, just need to do a single pass */
     if (missing_action == Fail)
     {
-        /* move to the left if it's l.e. than the split point */
+        /* move to the left if it's equal to the chosen category */
         for (size_t row = st; row <= end; row++)
         {
             if (x[ix_arr[row]] == split_categ)
@@ -2166,7 +2202,7 @@ void divide_subset_split(size_t *restrict ix_arr, int x[], size_t st, size_t end
         split_ix = st;
     }
 
-    /* otherwise, first put to the left all l.e. and not NA, then all NAs to the end of the left */
+    /* otherwise, first put to the left all equal to chosen and not NA, then all NAs to the end of the left */
     else
     {
         for (size_t row = st; row <= end; row++)
